#ifndef _RDBITEM_H_
#define _RDBITEM_H_

// RdbItem is a any type of attribute that the rdb interface supports.
// Data from other databases is placed here to allow conformity.

// The conversion operators do there best to convert from the objects
// type to the requested type, if possible.
// If the conversion is not possible, the program will die on an assertion!

// note:
// I tried to make this via inheritance, but it didn't work out,
// because I wanted functions to be able to return a generic attribute
// and virtual functions can only return one type.


#include "String.h"
#include "List.h"

typedef long   RdbInt;
typedef double RdbReal;

class RdbItem
{
  public:

    enum ItemType {
	RDB_STRING = 1,
	RDB_INT,
	RDB_REAL,
	RDB_NULL,
	// add more types here if needed
    };

    RdbItem( const RdbInt val );
    RdbItem( const RdbReal val );
    RdbItem( const String& val );
    RdbItem( const char* val );
    RdbItem();			// Null
    RdbItem( const RdbItem& other );

    ~RdbItem();
    RdbItem& operator =( const RdbItem& other );

    ItemType Type() const;	// returns the type of the item

    // convert from generic field to specific:
    operator RdbInt() const;
    operator RdbReal() const;
    operator String() const;

    friend ostream& operator <<(ostream& outs, const RdbItem& item);

  private:

    ItemType type;
    union {
	RdbInt   intVal;
	RdbReal  realVal;
	String*  strPtr;
	// add more types here if needed
    };

    void Cleanup();		// cleans up any allocation, but leaves the
				// object in a consistent state
};

typedef List<RdbItem> RdbItemList;


#include "RdbItem.i"

#endif // _RDBITEM_H_
