//********************************
// JoinAttrs.C
//
// Ravi Murthy
//********************************

#ifdef    WITH_PERSISTENCE
#include "JoinAttrs.h"

int hash_fn(int *start, int length);

// Constructor
// Initializes private members.
Cor_JoinAttrs::Cor_JoinAttrs(int *attrlist, int num)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    num_attrs = num;
    attrs = new int[num];

    for (int i = 0; i<num; i++)
        attrs[i] = attrlist[i];

}

// *************************************************************************
// This function first extracts the join attribute values from the
// binding enviroment. Then, it concatenates them into a single string.
// This is hashed and the value returned.
int Cor_JoinAttrs::hash_value(Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
//**** This is not quite right. Should first find the length
    const int MAXLEN = 100;
    char *curr_buf = new char[MAXLEN];
    char *curr_ptr = curr_buf;

    for (int i = 0; i< num_attrs; i++) {
        Cor_CArg arg = ((env->lookup(attrs[i])).expr);

        switch (arg.Kind()) {
   
        int    tmp_int;
        double tmp_double; 

        case  INT_ARG:
            tmp_int = arg.NumIval();
            memcpy(curr_ptr, &tmp_int, sizeof(int));
            curr_ptr += sizeof(int);
            break;
                
        case  DOUBLE_ARG: 
            tmp_double = arg.NumDval();
            memcpy(curr_ptr, &tmp_double, sizeof(double));
            curr_ptr += sizeof(double);
            break;
    
        case TINY_SYM_ARG: 
        case SYMBOL_ARG: 
            memcpy(curr_ptr, arg.symbol(), strlen(arg.symbol()));
            curr_ptr += strlen(arg.symbol());
            break;
                
        default:  
            cerr << "Unknown type of attribute\n";
        }
    }

    int hashval = hash_fn((int *)curr_buf, curr_ptr - curr_buf);
    delete curr_buf;
    return hashval;
}

//***************************************************************
// This function checks if the join attributes of the two envs are
// identical, and if so, superimposes bindings in env2 on env1. 
// Thus, if it returns TRUE, env1 will be the result of the JOIN
// of the two tuples.
int Cor_JoinAttrs::compare(Cor_BindEnv* env1, Cor_BindEnv *env2,
        Cor_ArgList& args1, Cor_ArgList& args2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i = 0;
    while ( (i < num_attrs)
        && ((env1->lookup(attrs[i]).expr) == (env2->lookup(attrs[i]).expr)))
        i++;

    if (i < num_attrs)
        return 0;       // Join values not identical

    // Now unify the bindings

    Cor_Term term;
    Cor_CArg *var2;
    for (i = 0; i < args2.count(); i++) {
        var2 = &(args2[i]);
        term = env2->lookup(var2->var_num());
        Cor_unify_binding(env1, var2->var_num(), term);
    }

    return 1;
}

//*****************************************************************


#define MULTIPLIER      39989
#define INCREMENT       13849
#define MAXHASH         32767
#define HASHVAL(x)         ( ((MULTIPLIER * (x)) + INCREMENT) % MAXHASH)

/////////////////////////////////////////////////////////////////////////
// hash_fn :  Hashes the data pointed to by start to an integer.       //
//            "length" is the length of the data.                      //
//                                                                     //
// Algo : the data is broken up into chunks of 4 bytes (and possible   //
//        one chunk with odd number of bytes). Each of the chunks are  //
//        ex-ored. This ex-ored value is then passed through a simple  //
//        random number generator "RAND" and the value returned is the //
//        hashed value of the data.                                    //
/////////////////////////////////////////////////////////////////////////
int hash_fn ( 
              int *start,                  // address of the key
              int length                   // bytes in the key
            )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    register unsigned int hash_base = 0;
    register int        k;

    for (k = (length >> 2);  k;  --k)   // get the full words
        hash_base ^= *start++;          // ex-or them 

    switch ( length & 03 )              // get the odd bytes
    {
    case 0:
        k = 0;  break;
    case 1:
        k = *start & 0x000000ff;  break;
    case 2:
        k = *start & 0x0000ffff;  break;
    case 3:
        k = *start & 0x00ffffff;
    }

    hash_base ^= k;

    return ( HASHVAL(hash_base) );
}

// *************************************************************************
#endif // WITH_PERSISTENCE
