//************************************
// Partition.C
//
// Ravi Murthy
//************************************

#ifdef    WITH_PERSISTENCE
#include <assert.h>
#include "Partition.h"

// Constructor
Cor_Partition::Cor_Partition(int num)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    num_part = num;
    parts = (Cor_PersistentRel **)new (Cor_PersistentRel*) [num_part];
    for (int i = 0; i < num_part; i++)
        parts[i] = NULL;
}

// Destructor
Cor_Partition::~Cor_Partition()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Dellocates the temporary partitions
    for (int i = 0; i < num_part; i++)
        delete parts[i];

    delete parts;
}

// *************************************************************************
// Based on the number of partitions previously declared, it creates
// as many temporary exodus relations.  Then, it scans the outer relation.
// For every tuple retrieved, it hashes it on the join attributes. Based
// on the hash value, it inserts the tuple in the appropriate relation.

int Cor_Partition::FormPartitions(Cor_TupleIterator* it, Cor_JoinAttrs *attr)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PersistentRel *rel = (Cor_PersistentRel *)(it->relation);

    int error;
    // create exodus relations
    for (int i = 0; i < num_part; i++) {
        // The name of the relation
        char *name = new char(strlen(rel->name.symbol()) + 22);
        sprintf(name, "%s#%d", rel->name.symbol(), i);
    
        // The hashval is given as 0 ...is that right?
        Cor_CArg *sym = new Cor_CArg(name);
        delete [] name;
    
        // Create a file desc
        Cor_FileDescStruct *fd = new Cor_FileDescStruct();
        fd->schema = (rel->file_desc()).schema;
        int e = sm_CreateFile(SM.bufGroup(),
                  SM.tempvolid(),
                  &(fd->fid));
        if (ErrorCheck(e, "sm_CreateFile"))
            return 0;

        fd->num_indexes = 0;
        fd->count = 0;
        fd->duplicates = !Cor_exEnv->C_check_subsum_default;

        parts[i] = new Cor_PersistentRel(*sym, rel->arity(), error, fd);
    }
    
    int hashval;
    it->reset();

    it->get_next_tuple();
    while (!(it->no_match())) {
        hashval = attr->hash_value(it->bindenv) % num_part;

        // Have to insert tuple in hashval th relation
        parts[hashval]->insert_new(it->arg_list, it->bindenv);

        it->get_next_tuple();
    }

    return 1;
}


// Creates a iterator over the specified partition.
Cor_TupleIterator *Cor_Partition::getIterator(int num, Cor_ArgList& args,
        Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (new Cor_TupleIterator(parts[num], args, env));
}

// *************************************************************************
#endif // WITH_PERSISTENCE
