
%{
#include "Globals.h"
#include "Scanner.h"
#include "CoralError.h"

#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "ArgList.h"
#include "CArg.h"

void yyerror(char *s);

EXTERN int yylex(YYSTYPE *lvalp);
EXTERN int yyparse();

Cor_CArg *Cor_retCArg;        // The result is returned in this
int       Cor_isquery;        // Is the result a query?

            // For table files - Info about the relation in progress
Cor_CArg *Cor_RelName = NULL;     // The relation's name
int       Cor_RelArity = 0;       // and its arity.

            // The variable names we've seen in this fact/arg
            // Their position indicates the variable id number to assign
Cor_CArgStack Cor_var_names;
int           Cor_num_vnames = 0;

            // Keep an _empty_ CArgStack here, you'll likely reuse it
Cor_CArgStack *Cor_stk_hook = NULL;

    // Redefine the parser to use this "lexer" so that we can switch
    // scanners (aka lexers) depending who is one the top of the stack.
    // The function will actually be called as:
    //     (parseEnv->lexer)(&yylval)
    // assuming YYPURE is defined and YYLSP_NEEDED is not.
    // This is done here to get around bison.simple's definition of YYLEX.
#undef    yylex
#define   yylex    (Cor_parseEnv->lexer)


%}

// This next line makes the parser re-entrant.
%pure_parser

%type   <carg>      arg arg_fact func_arg
%type   <ca_list>   arglist opt_arglist
%type   <carg>      member_list

// These tokens must stay the same as the ones in Gram.y, since we
// are using the same lexer! - SPF
%token  CONS NIL
%token  APPEND UPDATE ASSIGN DELETE
%token  MARK LPAREN RPAREN COMMA COLON SLASHSLASH PERIOD LBRACE RBRACE ARROW 
%token  SIPS LBRACKET RBRACKET UNDERSCORE VBAR PERCENT NOT QUESTION DOLLAR TRUE
%token  SUM PROD COUNT CMIN CMAX AVG ANY DISTINCT COUNT_D SUM_D PROD_D AVG_D
%token  DEL START_C_MODULE END_C_MODULE MODULE EXPORT ALIAS END_MODULE 
%token  AGG_SELECTOR AGG_SELECTOR_MANY_TO_ONE AGG_SELECTOR_MANY_TO_MANY
%token  PLUS MINUS STAR SLASH EQUAL NOTEQUAL LESS GREATER LESSEQUAL GREATEREQUAL
%token  LSHIFT RSHIFT BITXOR BITAND BITNOT BITOR MOD
%token  AT CUT DOUB_AT
%token  <carg> FSYMBOL SYMBOL VAR OP RELOP 
%token  <carg> NUM
%left   BITOR
%left   BITXOR
%left   BITAND
%left   LSHIFT RSHIFT
%left   PLUS MINUS
%left   STAR SLASH SLASHSLASH MOD
%right  BITNOT

%%

    // The seperator needs to come first so that the look-ahead
    // token is not another arg when its not there
fact_or_arg:    opt_seper arg
        {
            Cor_retCArg = $2;
            Cor_isquery = 0;

              // We only needed the names to get the correct variable #'s
            Cor_num_vnames = Cor_var_names.count();
            Cor_var_names.empty();
            YYACCEPT;
        }
    |         opt_seper QUESTION arg
        {
            Cor_retCArg = $3;
            Cor_isquery = 1;

              // We only needed the names to get the correct variable #'s
            Cor_num_vnames = Cor_var_names.count();
            Cor_var_names.empty();
            YYACCEPT;
        }

    |    opt_period    /* -empty file, or at end of fact file */
        {
            Cor_retCArg = NULL;
            Cor_isquery = 0;
            if (Cor_RelName)
                Cor_RelName->destroy();
            Cor_RelName = NULL;
            Cor_RelArity = 0;

            Cor_num_vnames = Cor_var_names.count();
            Cor_var_names.empty();
            YYACCEPT;
        }

    |    DOUB_AT  SYMBOL  NUM
        {
#ifdef    FIND_PARSE_LEAK
            Cor_CArg *tmp = $2->dup();
            delete $2;
            $2 = tmp;
            tmp = $3->dup();
            delete $3;
            $3 = tmp;
#endif
            // For table files - Info about the relation in progress
            if (Cor_RelName)
                Cor_RelName->destroy();

                // The name of the relation
            Cor_RelName = $2;

                // and its arity
            if ( ($3)->is_int() )
                Cor_RelArity = ($3)->NumIval();
            else
                Cor_RelArity = 0;

            $3->destroy();

              // return anyway to keep it simple
            Cor_retCArg = NULL;
            Cor_isquery = 0;
            YYACCEPT;
        }
;

opt_period:    PERIOD
    |        /* Nothing */
;

opt_seper:    PERIOD
    |         COMMA
    |        /* Nothing */
;

arglist:     arglist COMMA arg
        {
            ($1)->push($3);
        }

    |    arg
        {
            if (Cor_stk_hook) {
                $$ = Cor_stk_hook;
                Cor_stk_hook = NULL;
            } else 
                $$ = new Cor_CArgStack();
            ($$)->push($1);
        }
;

opt_arglist:    arglist
    |   /* Nothing */
        {
              // Actually allocate one so everyone can see its empty
            if (Cor_stk_hook) {
                $$ = Cor_stk_hook;
                Cor_stk_hook = NULL;
            } else 
                $$ = new Cor_CArgStack();
        }
;

arg:     arg_fact
    |    LPAREN arg RPAREN
        {
            $$ = $2;
        }

    |    MINUS arg
        {
            // They better have meant unary minus, no expressions
            // are allowed.
            $$ = $2;
            if (($$)->is_int()) {
                ($$)->set_num( - ($$)->NumIval() );
            } else if (($$)->is_double()) {
                ($$)->set_num( - ($$)->NumDval() );
            } else {
                *(Cor_exEnv->err_strm)
                << "Warning: Unary minus applied to a non-number (ignored).\n";
            }
        }
;

func_arg:   FSYMBOL LPAREN opt_arglist RPAREN
        {
#ifdef    FIND_PARSE_LEAK
            Cor_CArg *tmp = $1->dup();
            delete $1;
            $1 = tmp;
#endif
            $$ = Cor_create_functor($3, $1, NULL);
            $1->destroy();
            if (Cor_stk_hook)
                delete $3;
            else {
                Cor_stk_hook = $3;
                Cor_stk_hook->empty();
            }
        }
;

arg_fact:    SYMBOL
        {
#ifdef    FIND_PARSE_LEAK
          $$ = $1->dup();
          delete $1;
#endif
        }
    |        NUM
        {
#ifdef    FIND_PARSE_LEAK
          $$ = $1->dup();
          delete $1;
#endif
        }
    |        VAR
        {
            int vn = Cor_varname_enter(&Cor_var_names, $1);
            $$ = new Cor_CArg();
            $$->set_var(vn);
        }

    |        func_arg
    |        LBRACKET arglist VBAR arg RBRACKET
        {
            $$  = Cor_create_list($2, $4); 
            if (Cor_stk_hook)
              delete $2;
            else {
              Cor_stk_hook = $2;
              Cor_stk_hook->empty();
            }
        }

    |        LBRACKET opt_arglist RBRACKET
        {
 
            $$ = Cor_create_list($2, NULL);
            if (Cor_stk_hook)
               delete $2;
            else {
               Cor_stk_hook = $2;
               Cor_stk_hook->empty();
            }
        }

    |    CONS LPAREN arg COMMA arg RPAREN
        {
            $$ = new Cor_CArg($3,$5);
            //$3->destroy();
            //$5->destroy();
        }

    |    NIL
        {
            $$ = new Cor_CArg(NIL_ARG);
        }

    |       LBRACE member_list RBRACE
        {
            $$ = $2;
        } 
;

member_list:    member_list COMMA arg_fact 
        {
            Cor_insert_into_set($1, $3);
            $3->destroy();
        }

    |       arg_fact
        {
            $$ = Cor_create_nested_multiset();
            Cor_insert_into_set($$, $1);
            $1->destroy();
        }

    |       /* Nothing - allows the empty set {} */
        {
            $$ = Cor_create_nested_multiset();
        }
;

%%

// *************************************************************************
void yyerror(char *s)
{
    *(Cor_exEnv->err_strm) << s << " line "
            << Cor_parseEnv->cur_line_num << ":\n";

    if (Cor_exEnv->C_interrupt_raised > 0) {

        *(Cor_exEnv->err_strm)
               << "CORAL:: Interrupt occurred during execution!:";
        *(Cor_exEnv->err_strm) << "resetting state !\n";
        *(Cor_exEnv->err_strm) << "Type a period to continue.\n" << flush;

          // abort a consult that is halfway thru
        Cor_parseStack->abort_file();
        Cor_exEnv->C_interrupt_raised = 0;
    }
}

// *************************************************************************
