// *************************************************************************
//    Predicate.C
// *************************************************************************

#include  <string.h>
#include "Predicate.h"

#include "ParserStruct.h"
#include "Globals.h"
#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include <assert.h>
#include "Relation.h"
#include "SccInfo.h"
#include "Execute.h"
#endif

// *************************************************************************
Cor_Predicate::Cor_Predicate()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    arity = 0;
    clauses = NULL;
    clause_tail = NULL;
    work = NULL;
    allprednext = Cor_parserStruct->AllPredicates;
    ext_magic = 0;

    Cor_parserStruct->AllPredicates = this;
}

// *************************************************************************
Cor_Predicate::Cor_Predicate(Cor_ParserStruct *Cor_parserStruct,
        Cor_CArg& _name, int _arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    name = _name;
    arity = _arity;
    clauses = NULL;
    clause_tail = &clauses;
    work = NULL;
    allprednext = Cor_parserStruct->AllPredicates;
    ext_magic = 0;

    Cor_parserStruct->AllPredicates = this;
}

// *************************************************************************
Cor_Predicate::~Cor_Predicate()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    // Don't delete anything here, a Predicate is just a container
    // for pointers to things that belong to someone else.
}

// *************************************************************************
//  Adds a rule (a.k.a. clause) to the list of rules defining a predicate.
void Cor_AddClauseToPredicate(Cor_Predicate *predicate, Cor_Clause *clause)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    clause->pnext = *predicate->clause_tail;
    *predicate->clause_tail = clause;
    predicate->clause_tail = &clause->pnext;
}

// *************************************************************************
// Used to determine if the predicate is a magic predicate.  This may not be
// the right way to do it. Perhaps a subclass called MagicPredicate should
// be defined --- PS

int Cor_is_magic(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *str_name = name.symbol();

    if (str_name) {
        int len = strlen(str_name);
        if (len > 1) {
            return((str_name[0] == 'm') && (str_name[1] == '_'));
        }
    }
    
    return(0);
}

// *************************************************************************
int Cor_is_mp_done(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *str_name = name.symbol();

    if (str_name) {
        int len = strlen(str_name);

        if (len > 4) {
            return (strncmp(str_name, "done_", 5) == 0);
        }
    }

    return(0);
}

// *************************************************************************
int Cor_is_supp(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *str_name = name.symbol();

    if (str_name) {

        int len = strlen(str_name);

        if (len > 3) {
            return (strncmp(str_name,"sup_",4) == 0);
        }
    }

    return(0);
}

#ifdef ASYN_SOCK_REL
// *************************************************************************

int Cor_Predicate::asyn_sock()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation * rel = Cor_find_external_relation(name, arity);
    if (rel && !rel->_delta_done)
        return 1;
    else {
        char tmp[100];
        strcpy(tmp, name.symbol());
        char *ptr = strrchr(tmp, '_');

        if (ptr) {
            int is_adorn = 1;
            for (char *ptr3=ptr+1; *ptr3; ptr3++)
                if ((*ptr3 != 'b') && (*ptr3 != 'f')) {
                    is_adorn = 0;
                    break;
                }

            if (is_adorn)
                *ptr = '\0'; // get rid of the trailing part after '_'

            rel = Cor_find_external_relation(Cor_CArg(tmp), arity);
            if (rel && !rel->_delta_done)
                return 1;
        }
    }
    return 0;
}
#endif

#ifdef ASYN_SOCK_REL
int Cor_Predicate::delta_done()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Relation *rel = Cor_find_local_relation(name, arity);
    assert(rel);
    return rel->_delta_done;
}
#endif
