//************************************************************************
//    IO_Solvers.C
// *************************************************************************

#include <string.h>
#include <ctype.h>

#include "IO_Solvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "BuiltinUtil.h"
#include "Util.h"
#include "Database.h"
#include "Suspend.h"
#include "Suspension.h"

#include "Debug.h"

// *************************************************************************

#define TMP_CONSULT_FILE "/tmp/__consult.%s"

#define NOT_DEAD    10000

int Cor_reading_table = 0;

// *************************************************************************

#ifdef    SAVE_INTER
static int unlink(char *path)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    static cnt = 0;
    char cmmd[100];
    *cerr_strm << "Not unlinking " <<  path << " (moving to corsave." <<  cnt << ")\n";
    sprintf(cmmd, "/bin/mv %s /tmp/corsave.%d\n", path, cnt++);
    system(cmmd);
    return (0);
}

#endif   SAVE_INTER

// *************************************************************************
// cpp predefines certain names, and replaces them.  In order to avoid this,
// put any name you think might be a predefined name into this list.  Look
// at the man page for cpp on each host machine.  This is a very partial
// list so far.

#if defined(HPUX) || defined(LINUX)
#define COR_CPP_OPTS \
    "-P -Umips -Usparc -Uhp -Uultrix -Usun -Uunix -Uhpux -Ulinux"
#else
#define COR_CPP_OPTS \
    "-P -B -Umips -Usparc -Uhp -Uultrix -Usun -Uunix -Uhpux -Ulinux"
#endif

// *************************************************************************

static void user_continue();

extern "C" int CArg_parse(void);

// *************************************************************************
Cor_BindEnv *Cor_EmptyPrintSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_TuplePrintSolver(Cor_BuiltinRelation &,
                                  Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!Cor_exEnv->C_verbose_answer_default)
        return(iterator.bindenv);

    int count = iterator.arg_list.count();

    if (count > 0) {
        *(Cor_exEnv->out_strm) << " (";

     /*
        Cor_ArgList *copy_args = new Cor_ArgList(count);
        Cor_CopyArgs(copy_args, &iterator.arg_list,
                   iterator.bindenv, count, NULL);

        FOR_EACH_ARG(arg, (*copy_args)) {
            arg->printon(Cor_exEnv->out_strm);
            if (--count)
                *(Cor_exEnv->out_strm) << ", ";
        }
        END_EACH_ARG;

        delete copy_args;
    */

        iterator.arg_list.print(iterator.bindenv,Cor_exEnv->out_strm);

        *(Cor_exEnv->out_strm) << ")\n";
    } else {
        if (Cor_exEnv->C_verbose_answer_default)
            *(Cor_exEnv->out_strm) << "yes.\n";
    }

    Cor_exEnv->out_strm->flush();

    if (Cor_exEnv->C_interactive_mode_current)
        user_continue();

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_BindingsPrintSolver(Cor_BuiltinRelation &,
                                     Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    
    int count = iterator.arg_list.count();

    if (count > 0) {

        Cor_ArgList *copy_args = new Cor_ArgList(count);
        Cor_CopyArgs(copy_args, &iterator.arg_list,
                     iterator.bindenv, count, NULL);

        for (int i = 0; i < count;) {
            iterator.arg_list[i].printon(Cor_exEnv->out_strm);
            *(Cor_exEnv->out_strm) << "=";
            (*copy_args)[i].printon(Cor_exEnv->out_strm);

            if (++i < count)
                *(Cor_exEnv->out_strm) << ", ";
        }

        delete copy_args;
        *(Cor_exEnv->out_strm) << ".\n";
    } else {
        if (Cor_exEnv->C_verbose_answer_default)
            *(Cor_exEnv->out_strm) << "yes.\n";
    }

    *(Cor_exEnv->out_strm) << flush;

    if (Cor_exEnv->C_interactive_mode_current)
        user_continue();

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_LiteralPrintSolver(Cor_BuiltinRelation &,
                                  Cor_TupleIterator &iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int  j, count;
    Cor_Literal *query_literal;

    count = Cor_parserStruct->rhspreds.count();

    for (j=0; j < count; j++) {

        query_literal =
                (Cor_Literal *) Cor_parserStruct->rhspreds.get_ith(j);

        *(Cor_exEnv->out_strm) << query_literal->pred.symbol() << "(";

        query_literal->args.print(iterator.bindenv,Cor_exEnv->out_strm);

        if (j == count-1) 
            *(Cor_exEnv->out_strm) << ")";
        else
            *(Cor_exEnv->out_strm) << "), ";
    }

    *(Cor_exEnv->out_strm) << ".\n";

    Cor_exEnv->out_strm->flush();

    if (Cor_exEnv->C_interactive_mode_current)
        user_continue();

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_PrintSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int count = iterator.arg_list.count();

    if (count > 0) {
        Cor_ArgList *copy_args = new Cor_ArgList(count);
        Cor_CopyArgs(copy_args, &iterator.arg_list,
                     iterator.bindenv, count, NULL);

        FOR_EACH_ARG(arg, (*copy_args)) {
            arg->printon(Cor_exEnv->out_strm);
            if (--count)
                *(Cor_exEnv->out_strm) << " ";
        }
        END_EACH_ARG;

        delete copy_args;
    }
    Cor_exEnv->out_strm->flush();

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_PrintfSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int argcount = iterator.arg_list.count();

    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "PrintfSolver");
        return (iterator.bindenv);
    }
    Cor_ArgList *copy_args = new Cor_ArgList(argcount);
    Cor_CopyArgs(copy_args, &iterator.arg_list,
                 iterator.bindenv, argcount, NULL);

    Cor_Term term1(copy_args->arg(0), NULL);
    if (!term1.expr.is_symbol()) {
        CORAL_error(COR_BAD_ARG_TYPE, "arg 1 isn't a format string",
                    "PrintSolver");
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        iterator.set_no_match();
        return (iterator.bindenv);
    }

    char *format_str = new char[200];
    char *temp1;
    char *temp2 = term1.expr.symbol();
    int count = 1;
    for (; *temp2;) {
        if (*temp2 == '\0') {
          error:break;
        } else if (*temp2 == '%') {
            temp1 = format_str;
            *temp1 = '%';
            temp1++;

            temp2++;
            if (*temp2 == '-') {
                *temp1 = '-';
                temp1++;
                temp2++;
            }
            while ((*temp2 >= '0') && (*temp2 <= '9')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            if (*temp2 == '.') {
                *temp1 = '.';
                temp1++;
                temp2++;
            }
            while ((*temp2 >= '0') && (*temp2 <= '9')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            if ((*temp2 == 'h') || (*temp2 == 'l')) {
                *temp1 = *temp2;
                temp1++;
                temp2++;
            }
            switch (*temp2) {
            case 'd':
            case 'i':
            case 'u':
            case 's':
            case 'f':
            case 'e':
            case 'E':
            case 'g':
            case 'G':
                *temp1 = *temp2;
                temp1++;
                temp2++;
                *temp1 = '\0';
                if (count >= argcount) {
                    *(Cor_exEnv->err_strm) << 
                                     "Not enough arguments to print.\n";
                    goto error;
                } {
                    Cor_Term term(copy_args->arg(count), NULL);
                    count++;
                    //term.expr.print(NULL, Cor_exEnv->out_strm, format_str);
                    term.expr.print(NULL, Cor_exEnv->out_strm);
                    temp1 = format_str;
                    break;
                }
            case '%':
                Cor_exEnv->out_strm->putch('%');
                break;
            }
        } else {
            Cor_exEnv->out_strm->putch(*temp2);
            temp2++;
        }
    }

    delete [] format_str;
    Cor_exEnv->out_strm->flush();

    delete copy_args;
    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_ConsultSolver(Cor_BuiltinRelation &,
                               Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char tmp_consult[MAX_PATH_LEN];

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << "Entering %s\n", "consult";
    }
#endif

    sprintf(tmp_consult, TMP_CONSULT_FILE, coral_pid);

    if (iterator.arg_list.count() < 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to consult needed",
                    "ConsultSolver");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    for (int j=iterator.arg_list.count()-1; j >= 0; j--) {

      Cor_Term term1(iterator.arg_list[j], iterator.bindenv);
      Cor_deep_dereference_term(term1);
  
      if (!term1.expr.is_symbol()) {
#ifdef    DEBUG
          cerr << "Arg given was '" << term1.expr << endl;
#endif
          CORAL_error(COR_BAD_ARG_TYPE, NULL, "ConsultSolver");
          term1.printon(Cor_exEnv->err_strm);
          *(Cor_exEnv->err_strm) << endl;
          iterator.set_no_match();
          continue;
      }
  
      char *file_name = term1.expr.symbol();
  
      if (! *file_name) {
             // Consult of "" is always successful 
          iterator.reset_no_match();
          continue;
      }
  
      if (Cor_tableFile(file_name)) {
          Cor_ReadTable(file_name, iterator);
          continue;
      } else {
          char term_str[200];

	  sprintf(term_str, "%scpp  %s %s > %s" , CPP_PATH,  COR_CPP_OPTS,
		  file_name, tmp_consult);
  
          if (system(term_str)) {
              *(Cor_exEnv->err_strm)
                   << "CORAL::error : Unable to execute the shell command\n";
              *(Cor_exEnv->err_strm) << term_str << endl;
              *(Cor_exEnv->err_strm) << "Aborting attempt to consult the file '"
                    << file_name << "'.\n";
              unlink(tmp_consult);
              iterator.set_no_match();
              continue;
          }
  
          if (!Cor_consultFile(tmp_consult, file_name)) {
              CORAL_error(COR_FILE_NOT_FOUND, file_name, "ConsultSolver");
              iterator.set_no_match();
              unlink(tmp_consult);
              continue;
          }
          unlink(tmp_consult);
      }
    }

    return(iterator.bindenv);
}

// *************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();  \
    return(NULL); \
}

// This routine takes the name of the file in, and processes it.
//
Cor_BindEnv *Cor_ReadTable(char *infile_nme, Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    if (!Cor_consultFile(infile_nme, infile_nme)) {
        CORAL_error(COR_FILE_NOT_FOUND, infile_nme, "ReadTable");
        iterator.set_no_match();
        return(iterator.bindenv);
    }

    extern Cor_CArg *Cor_retCArg;
    extern Cor_CArg *Cor_RelName;

    Cor_reading_table = 1;

    if (Cor_retCArg) {
        Cor_retCArg->destroy();
        Cor_retCArg = NULL;
    }

    if (Cor_RelName) {
        Cor_RelName->destroy();
        Cor_RelName = NULL;
    }

    // Now read in the lines using parseFact() and insert into the relation
    int cnt = 0;
    Cor_Relation *rel = NULL;
    Cor_Tuple *tuple = NULL;

    Cor_CArg prev_relnme;
    Cor_CArg relnme;
    int rel_arity = 0;
    int prev_arity = 0;

    tuple = Cor_parseFact(NULL, relnme, &rel_arity);

    while (tuple) {

        // If its a new relation find or create it
        if ((rel_arity != prev_arity) || (!relnme.equals(prev_relnme))) {

            rel = Cor_find_local_or_external_relation(relnme, rel_arity);
            prev_relnme = relnme;
            prev_arity = rel_arity;
        }

        // Check for completely ground (ie: valid) at the top-level
        for (j=0; j < rel_arity; j++) {
            if ((*(tuple->_args))[j].is_var()) {
                *(Cor_exEnv->err_strm) << "ReadTable: invalid fact: "
                                       << *tuple << " ignored.\n";
                delete tuple;
                break;
            }
        }

        if (j == rel_arity) {

        // Insert fact into Relation rel
#if    1
        if (! rel->insert_tuple(tuple))
            delete tuple;
#else
        // This leaks like crazy
        if (!rel->insert_new(tuple))
            delete tuple;
#endif
        }

        // Can't delete tuple, as the relation stores it
        // (and not just the args)

        if ((++cnt % NOT_DEAD) == 0) {
            *(Cor_exEnv->out_strm) << cnt << endl;
        }
        tuple = Cor_parseFact(NULL, relnme, &rel_arity);
    }

    Cor_reading_table = 0;

    return(iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match(); \
    if (outfile_nme && del_outfile) { \
       unlink(outfile_nme); \
       cout << "del_outfilename =" <<  del_outfile << endl; \
       delete outfile; \
       } \
    return(NULL); \
}

Cor_BindEnv *Cor_WriteTableSolver(Cor_BuiltinRelation &,
                                  Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Arglist can have one or two args. The first is the relation name,
    // and the second is the output file. If the second argument is not
    // provided, the output is sent to stdout

    ostream *outfile = NULL;
    char *outfile_nme = NULL;
    Cor_CArg outfle_carg;
    int  del_outfile = 1;

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << 
                 "Entering " <<  Cor_WriteTableSymbol->symbol() << endl;
    }
#endif

    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_functor()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad first argument to "
            << Cor_WriteTableSymbol->symbol() << " ="
            << term1 << endl;
        FAIL_RETURN;
    }

    Cor_CArg *fa = &(term1.expr);

    if (iterator.arg_list.arity() > 1) {
        Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term(term2);
        outfle_carg = term2.expr;
        outfile_nme = outfle_carg.symbol();
        *(cerr_strm) << "Output file name is: " << outfile_nme << endl;
    }

    Cor_Relation *printrel = find_relation(fa->functor(), fa->arity());
    if (!printrel) {
        *(Cor_exEnv->err_strm)
              << "CORAL :: Error - " << Cor_WriteTableSymbol->symbol()
              << ": relation " << fa->functor()
              << "(arity = " << fa->arity() << ") not found.\n";
        del_outfile = 0;
        FAIL_RETURN;
    }

    if (outfile_nme) {
        outfile = Cor_open_outfile(outfile_nme, ios::app);
    } else {
        outfile = Cor_exEnv->out_strm;
        del_outfile = 0;
    }

    if (!outfile) {
        *(Cor_exEnv->err_strm) << "Could not open file '"
                               << outfile_nme << "'\n";
        FAIL_RETURN;
    }

    *outfile << "\n@@ " <<  fa->functor() << " " <<  fa->arity() << "\n";

    printrel->Cor_Relation::print_facts(outfile,NULL,1);

    if (del_outfile)
        delete outfile;

    iterator.reset_no_match();
    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) delete tmp_file;     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

Cor_BindEnv *Cor_pipeInSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //    Arglist should have one arg, which is  a command to be run and its
    //    output used as input to the system

    ostream *tmp_file = NULL;
    char *tmp_file_name = NULL;

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm)
            << "Entering " << Cor_PipeInSymbol->symbol() << endl;
    }
#endif

    if (iterator.arg_list.arity() != 1) {
        *(Cor_exEnv->err_strm) << Cor_PipeInSymbol->symbol()
            << ": Number of arguments must be 1.\n";
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_symbol()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad first argument to "
            << Cor_PipeInSymbol->symbol() << " =" << term1 << endl;
        FAIL_RETURN;
    }
    char *command = term1.expr.symbol();

    char *term_str = new char[strlen(command) + 100];

    char name[100];
    tmp_file_name = Cor_mk_tmp_file_name(name);

    sprintf(term_str, "%s > %s", command, tmp_file_name);

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << Cor_PipeInSymbol->symbol()
                  << ": Executing command " << term_str << endl;
    }
#endif

    if (system(term_str)) {
        *(Cor_exEnv->err_strm) << 
               "CORAL :: Error -- Command \"" << term_str << "\" failed !!\n";
        delete[]term_str;
        unlink(tmp_file_name);
        FAIL_RETURN;
    }
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << Cor_PipeInSymbol->symbol()
            << " - redirecting input.\n";
    }
#endif

    if (!Cor_consultFile(tmp_file_name, tmp_file_name)) {
        CORAL_error(COR_FILE_NOT_FOUND, tmp_file_name, "pipeInSolver");
        iterator.set_no_match();
        unlink(tmp_file_name);
        delete[]term_str;
        return (iterator.bindenv);
    }
    delete[]term_str;
    unlink(tmp_file_name);
    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
#define FAIL_RETURN { \
    iterator.set_no_match();      \
    if (tmp_file) delete tmp_file;     \
    if (tmp_file_name) unlink(tmp_file_name); \
    return(NULL); \
}

Cor_BindEnv *Cor_pipeOutSolver(Cor_BuiltinRelation &,
                               Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Arglist should have two args.  The first a command to which the output
    // is given, the second argument a relation name.
    ostream *tmp_file = NULL;
    char *tmp_file_name = NULL;

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << "Entering "
            << Cor_PipeOutSymbol->symbol() << endl;
    }
#endif

    if (iterator.arg_list.arity() != 2) {
        *(Cor_exEnv->err_strm) << Cor_PipeOutSymbol->symbol()
            << ": Number of arguments must be 2.\n";
        FAIL_RETURN;
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_symbol()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad first argument to "
            << Cor_PipeOutSymbol->symbol() << " =" << term1 << endl;
        FAIL_RETURN;
    }
    Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term(term2);
    if (!term2.expr.is_functor()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad second argument to "
            << Cor_PipeOutSymbol->symbol() << " =" << term2 << endl;
        FAIL_RETURN;
    }
    Cor_CArg *fa = &(term2.expr);

    char name[100];
    tmp_file_name = Cor_mk_tmp_file_name(name);
    tmp_file = Cor_open_file(tmp_file_name);
    if (!tmp_file)
        FAIL_RETURN;

    Cor_Relation *printrel = find_relation(fa->functor(), fa->arity());
    if (!printrel) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error - " << Cor_PipeOutSymbol->symbol()
            << ": relation " << fa->functor()
            << "(arity = " << fa->arity() << ") not found.\n";
        FAIL_RETURN;
    }
#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm) << 
                   Cor_PipeOutSymbol->symbol() << " - printing facts..\n";
    }
#endif

    printrel->print_facts(tmp_file);
    delete tmp_file;
    tmp_file = NULL;

    char *command = term1.expr.symbol();
    char *term_str = new char[strlen(command) + 100];
    sprintf(term_str, "%s < %s", command, tmp_file_name);

#ifdef DO_TRACE
    if (Cor_exEnv->dbg_get_next) {
        *(Cor_exEnv->trace_strm)
           << Cor_PipeOutSymbol->symbol()
           << " - going to execute command: " << term_str << endl;
    }
#endif

    if (system(term_str)) {
        *(Cor_exEnv->err_strm) << 
               "CORAL :: Error -- Command \"" <<  term_str << "\" failed !!\n";
        delete [] term_str;
        FAIL_RETURN;
    }
    delete[]term_str;

    unlink(tmp_file_name);
    iterator.reset_no_match();

    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_BindEnv *Cor_ShellSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_symbol()) {
        *(Cor_exEnv->err_strm) << 
                "CORAL :: Error -- bad argument to shell =";
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    char *term_str = term1.expr.symbol();
    if (system(term_str)) {
        CORAL_error(COR_FAILED_COMMAND, term_str, "ShellSolver");
        iterator.set_no_match();
    }
    return (iterator.bindenv);
}

// *************************************************************************
static void user_continue()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char input[100];

    if (Cor_suspension->set > 0) {   // previous suspension command.

        if (Cor_suspension->set > 1)
            Cor_suspension->set--;   // generate the number of answers required.

        return;
    }

    if (Cor_suspension->set == 0)    // suspension according to page
        return;

    if (coralConfig->server_mode) {        // never gets here!
        //*(Cor_exEnv->out_strm) << "Next Answer ? (y/n/all/9)[y]\n";
        //Cor_parseEnv->print_prompt();

        *(Cor_exEnv->err_strm)
            << "SEVERE WARNING: INFORM CORAL PROGRAMMER!!!\n"
            << "user_continue: server, suspension.set = "
            << Cor_suspension->set << endl;
        return;
    } else
        *(Cor_exEnv->out_strm) << "\t... next answer ? (y/n/all)[y]";

    Cor_exEnv->in_strm->gets(input, 100);

    // Hack to get rid of excess \n
    if (input[0] == '\n')
        Cor_exEnv->in_strm->gets(input, 100);

    if ((input[0] == 'n') || (input[0] == 'N')) {
        // prevent any further computation
        Cor_exEnv->C_interrupt_raised = -1;
    } else if ((input[0] == 'a') || (input[0] == 'A')) {
        // prevent further answer prompting
        Cor_exEnv->C_interactive_mode_current = 0;
    } else if (isdigit(input[0]))
        Cor_suspension->set = 1 + atoi(input);        //suspended!

}

// *************************************************************************
