// *************************************************************************
//    InitBuiltins.C
// *************************************************************************

#include "InitBuiltins.h"

#include "CArg.h"
#include "BuiltinSymbols.h"
#include "Globals.h"
#include "DatabaseStruct.h"
#include "ModuleInfo.h"
#include "BindEnv.h"

#include "Operators.h"
#include "BuiltinRelation.h"
#include "BuiltinSetRelation.h"
#include "BuiltinCompareRelation.h"
#include "BuiltinUnaryRelation.h"
#include "BuiltinBinaryRelation.h"
#include "BuiltinTraceRelation.h"
#include "BuiltinDebugRelation.h"
#include "BuiltinIndexRelation.h"
#include "BuiltinFailRelation.h"
#include "BuiltinUserTupleRelation.h"
#include "Debug.h"
#include "SocketRelSolver.h"
#include "AsynSocketRelSolver.h"
#include "UrlRelSolver.h"

// *************************************************************************

#include "Solvers.h"
#include "DBSolvers.h"
#include "DebugSolvers.h"
#include "IO_Solvers.h"
#include "PersistSolvers.h"
#include "ServerSolvers.h"
#include "SetSolvers.h"

// *************************************************************************

extern Cor_ModuleInfo          *Cor_CurrModuleInfo;
extern Cor_BuiltinSetRelation   MemberBuiltin;

// *************************************************************************

extern Cor_BuiltinUnaryRelation  SumBuiltin;
extern Cor_BuiltinUnaryRelation  CountBuiltin;
extern Cor_BuiltinUnaryRelation  MinBuiltin;
extern Cor_BuiltinUnaryRelation  MaxBuiltin;
extern Cor_BuiltinUnaryRelation  AvgBuiltin;

extern Cor_BuiltinBinaryRelation AddBuiltin;
extern Cor_BuiltinBinaryRelation SubBuiltin;
extern Cor_BuiltinBinaryRelation MulBuiltin;
extern Cor_BuiltinBinaryRelation DivBuiltin;

extern Cor_BuiltinCompareRelation EquBuiltin;
extern Cor_BuiltinCompareRelation NeqBuiltin;
extern Cor_BuiltinCompareRelation LssBuiltin;
extern Cor_BuiltinCompareRelation GrtBuiltin;
extern Cor_BuiltinCompareRelation LeqBuiltin;
extern Cor_BuiltinCompareRelation GeqBuiltin;

extern Cor_BuiltinTraceRelation TraceBuiltin, UntraceBuiltin;
extern Cor_BuiltinDebugRelation SetBuiltin, ClearBuiltin, AssignBuiltin;
extern Cor_BuiltinRelation      DisplayBuiltin, HelpBuiltin, PrintfBuiltin;
extern Cor_BuiltinRelation      StartTimeRelation, DisplayTimeRelation;
extern Cor_BuiltinRelation      ConsultBuiltin;
extern Cor_BuiltinIndexRelation AddIndexBuiltin;
extern Cor_BuiltinFailRelation  FailBuiltin;

extern Cor_BuiltinSetRelation   ExtJoinBuiltin;
extern Cor_BindEnv *Cor_ExtJoinSolver(Cor_TupleIterator&);

// *************************************************************************

void Cor_InsertBuiltin(Cor_CArg& n, Cor_Relation *r)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!Cor_BuiltinDB->RelationTable)
        Cor_BuiltinDB->RelationTable = new Cor_SymTable();

    Cor_BuiltinDB->RelationTable->insert(n, (Pointer) r);
}

// *************************************************************************
void Cor_initBuiltins()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BuiltinRelation *temp;

    temp = new Cor_BuiltinRelation(-1, Cor_UrlRelSymbol, UrlRelSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_SocketRelSymbol, SocketRelSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_AsynSocketRelSymbol, AsynSocketRelSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_WhoSymbol, Cor_WhoSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_WhoamiSymbol, Cor_WhoamiSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_LoginSymbol, Cor_LoginSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_LogoutSymbol, Cor_LogoutSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_AnswerSymbol, Cor_AnswerSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_UserSymbol, Cor_UserSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_YieldSymbol, Cor_YieldSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_SetQidSymbol, Cor_SetQidSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_CollectionSymbol,
                                                     Cor_CollectionSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_ListSusSymbol, Cor_ListSusSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_ResumeSymbol, Cor_ResumeSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_SyncSymbol, Cor_SyncSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_KillSymbol, Cor_KillSolver);

    temp = new Cor_BuiltinRelation(-1, Cor_BreakSymbol, Cor_BreakSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_ClearBreakSymbol,
            Cor_ClearBreakSolver);

    temp = new Cor_BuiltinRelation(-1, Cor_PrintModSymbol,
            Cor_PrintModuleSolver);
    
    temp = new Cor_BuiltinSetRelation(-1, Cor_MemTupleSymbol,
            Cor_MemTupleSolver);

    temp = new Cor_BuiltinRelation(-1, Cor_SortSymbol, Cor_SortRelSolver);

    temp = new Cor_BuiltinSetRelation(-1, Cor_MemberSymbol, Cor_MemberSolver); 
    temp = new Cor_BuiltinSetRelation(1, Cor_CallSymbol, Cor_CallSolver);
    temp = new Cor_BuiltinRelation(2, Cor_UnivSymbol, Cor_UnivSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_ArgsSymbol, Cor_ArgsSolver);

    temp = new Cor_BuiltinSetRelation(2, Cor_SubsetSymbol, Cor_SubsetSolver);
    temp = new Cor_BuiltinSetRelation(3, Cor_UnionSumSymbol, Cor_UnionSolver);
    temp = new Cor_BuiltinSetRelation(2, Cor_UnionSumDSymbol, Cor_UnionDSolver);
    temp = new Cor_BuiltinSetRelation(3, Cor_UnionMaxSymbol,
            Cor_MultiUnionSolver);
    temp = new Cor_BuiltinSetRelation(2, Cor_UnionMaxDSymbol,
            Cor_MultiUnionDSolver);
    temp = new Cor_BuiltinSetRelation(3, Cor_InterSymbol,
            Cor_IntersectionSolver);
    temp = new Cor_BuiltinSetRelation(2, Cor_InterDSymbol,
            Cor_IntersectionDSolver);
    temp = new Cor_BuiltinSetRelation(3, Cor_DifferenceSymbol, Cor_DiffSolver);
    temp = new Cor_BuiltinSetRelation(2, Cor_DifferenceDSymbol,
            Cor_DiffDSolver);
    temp = new Cor_BuiltinSetRelation(2, Cor_MakeSetSymbol, Cor_MakeSetSolver);
    temp = new Cor_BuiltinSetRelation(1, Cor_MakeSetDSymbol,
            Cor_MakeSetDSolver);
    temp = new Cor_BuiltinUserTupleRelation(1, Cor_CreateSetSymbol,
            Cor_CreateSetSolver);
    temp = new Cor_BuiltinUserTupleRelation(3, Cor_AddElemSymbol,
            Cor_AddElemSolver);
    temp = new Cor_BuiltinUserTupleRelation(2, Cor_AddElemDSymbol,
            Cor_AddElemDSolver);
    temp = new Cor_BuiltinUserTupleRelation(2, Cor_AbsOpSymbol, Cor_AbsSolver);
    temp = new Cor_BuiltinUserTupleRelation(2, Cor_CeilOpSymbol,
            Cor_CeilSolver);
    temp = new Cor_BuiltinUserTupleRelation(2, Cor_FloorOpSymbol,
            Cor_FloorSolver);

    temp = new Cor_BuiltinUserTupleRelation(3, Cor_FunctorSymbol, Cor_FuncOp);
    temp = new Cor_BuiltinUserTupleRelation(3, Cor_ArgSymbol, Cor_ArgOp);
    temp = new Cor_BuiltinUserTupleRelation(3, Cor_PowSymbol, Cor_PowSolver);

    temp = new Cor_BuiltinUserTupleRelation(2, Cor_LengthSymbol,
            Cor_LengthSolver);
    temp = new Cor_BuiltinUserTupleRelation(3, Cor_ExtractSymbol,
            Cor_ExtractSolver);
    temp = new Cor_BuiltinUserTupleRelation(3, Cor_SubstringSymbol,
            Cor_SubstringSolver);


    temp = new Cor_BuiltinBinaryRelation(3, Cor_LShiftOpSymbol, Cor_LShiftOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_RShiftOpSymbol, Cor_RShiftOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_BitandOpSymbol, Cor_BitandOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_BitorOpSymbol, Cor_BitorOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_BitxorOpSymbol, Cor_BitxorOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_ModOpSymbol, Cor_ModOp);

    temp = new Cor_BuiltinBinaryRelation(3, Cor_AddOpSymbol, Cor_AddOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_SubOpSymbol, Cor_SubOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_MulOpSymbol, Cor_MulOp);
    temp = new Cor_BuiltinBinaryRelation(3, Cor_DivOpSymbol, Cor_DivOp);
    
    temp = new Cor_BuiltinUnaryRelation(2, Cor_SumOpSymbol, Cor_SumOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_ProdOpSymbol, Cor_ProdOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_CountOpSymbol, Cor_CountOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_MaxOpSymbol, Cor_MaxOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_MinOpSymbol, Cor_MinOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_AvgOpSymbol, Cor_AvgOp);
    temp = new Cor_BuiltinUnaryRelation(2, Cor_StdDevOpSymbol, Cor_StdDevOp);

    temp = new Cor_BuiltinCompareRelation(2, Cor_EquOpSymbol, Cor_EquOp);
    temp = new Cor_BuiltinCompareRelation(2, Cor_NeqOpSymbol, Cor_NeqOp);
    temp = new Cor_BuiltinCompareRelation(2, Cor_LssOpSymbol, Cor_LssOp);
    temp = new Cor_BuiltinCompareRelation(2, Cor_GrtOpSymbol, Cor_GrtOp);
    temp = new Cor_BuiltinCompareRelation(2, Cor_LeqOpSymbol, Cor_LeqOp);
    temp = new Cor_BuiltinCompareRelation(2, Cor_GeqOpSymbol, Cor_GeqOp);

    temp = new Cor_BuiltinFailRelation(0, Cor_FailSymbol);
    temp = new Cor_BuiltinRelation(0, Cor_CutSymbol, Cor_CutSolver);
    temp = new Cor_BuiltinRelation(0, Cor_QuitSymbol, Cor_QuitSolver);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_ExplainOnSymbol);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_ExplainOffSymbol);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_TraceSymbol);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_UntraceSymbol);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_ProfileOnSymbol);
    temp = new Cor_BuiltinTraceRelation(-1, Cor_ProfileOffSymbol);

    temp = new Cor_BuiltinDebugRelation(-1, Cor_SetSymbol);
    temp = new Cor_BuiltinDebugRelation(-1, Cor_ClearSymbol);
    temp = new Cor_BuiltinDebugRelation(2, Cor_AssignSymbol);

    temp = new Cor_BuiltinRelation(-1, Cor_HelpSymbol, Cor_HelpSolver);
    temp = new Cor_BuiltinRelation(1, Cor_ConsultSymbol, Cor_ConsultSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_DisplaySymbol, Cor_PrintSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_PrintfSymbol, Cor_PrintfSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_PrintTupleSymbol,
            Cor_TuplePrintSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_PrintBindingSymbol,
            Cor_BindingsPrintSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_PrintEmptySymbol,
            Cor_EmptyPrintSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_PrintLiteralSymbol,
            Cor_LiteralPrintSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_StartTimerSymbol,
            Cor_StartTimeSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_DisplayTimerSymbol,
            Cor_EndTimeSolver);
    temp = new Cor_BuiltinUserTupleRelation(1, Cor_CpuTimeSymbol,
            Cor_CputimeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_ShellSymbol, Cor_ShellSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_ListRelsInDBSymbol, Cor_RelDisplay);
    temp = new Cor_BuiltinRelation(1, Cor_PipeInSymbol, Cor_pipeInSolver);
    temp = new Cor_BuiltinRelation(2, Cor_PipeOutSymbol, Cor_pipeOutSolver);
    temp = new Cor_BuiltinRelation(2, Cor_WriteTableSymbol,
            Cor_WriteTableSolver);

    temp = new Cor_BuiltinRelation(-1, Cor_ShowMemSymbol, Cor_ShowMemSolver);

    temp = new Cor_BuiltinRelation(1, Cor_MakeIndexSymbol, Cor_addIndexSolver1);
    temp = new Cor_BuiltinRelation(2, Cor_AliasSymbol, Cor_AliasSolver);
    temp = new Cor_BuiltinRelation(0, Cor_DisplayDefaultsSymbol,
            Cor_DisplayDefaultsSolver);
    temp = new Cor_BuiltinRelation(0, Cor_ResetDefaultsSymbol,
            Cor_ResetDefaultsSolver);

    temp = new Cor_BuiltinRelation(1, Cor_CreateDBSymbol, Cor_CreateDBSolver);
    temp = new Cor_BuiltinRelation(1, Cor_DestroyDBSymbol, Cor_DestroyDBSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_CurrentDBSymbol,
            Cor_CurrentDBSolver);
    temp = new Cor_BuiltinRelation(-1, Cor_CoralRootSymbol,
            Cor_CoralRootSolver);
    temp = new Cor_BuiltinRelation(2, Cor_SaveDBSymbol, Cor_SaveDBSolver);
    temp = new Cor_BuiltinRelation(2, Cor_RenameDBSymbol, Cor_RenameDBSolver);
    temp = new Cor_BuiltinRelation(2, Cor_RenameRelSymbol, 
  	    Cor_RenameRelSolver); 
    temp = new Cor_BuiltinRelation(1, Cor_ClearDBSymbol, Cor_ClearDBSolver);
    temp = new Cor_BuiltinRelation(1, Cor_ListDBSymbol, Cor_ListDBSolver);
    temp = new Cor_BuiltinRelation(2, Cor_OpenRelInDBSymbol, Cor_OpenRelSolver);
    temp = new Cor_BuiltinRelation(1, Cor_CloseRelSymbol, Cor_CloseRelSolver);
    temp = new Cor_BuiltinRelation(1, Cor_ClearRelSymbol, Cor_ClearRelSolver);
    temp = new Cor_BuiltinRelation(1, Cor_CopyRelToDBSymbol, Cor_CopyRelSolver);
    temp = new Cor_BuiltinRelation(1, Cor_AddRelToDBSymbol, Cor_AddRelSolver);
    temp = new Cor_BuiltinRelation(1, Cor_InsertFactSymbol,
            Cor_InsertFactSolver);
    temp = new Cor_BuiltinRelation(1, Cor_DeleteFactSymbol,
            Cor_DeleteFactSolver);
    temp = new Cor_BuiltinRelation(1, Cor_HistorySymbol, Cor_HistorySolver);

    temp = new Cor_BuiltinRelation(1, Cor_IsConstantSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsVarSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsListSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsFunctorSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsNumSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsStringSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsIntSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsLongSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsShortSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsFloatSymbol, ArgTypeSolver);
    temp = new Cor_BuiltinRelation(1, Cor_IsDoubleSymbol, ArgTypeSolver);

    temp = new Cor_BuiltinRelation(1, Cor_CounterSymbol, Cor_CounterSolver);
    temp = new Cor_BuiltinRelation(1, Cor_ResetCounterSymbol,
            Cor_ResetCounterSolver);


      // Install schema, and bulkload builtin relations
    temp = new Cor_BuiltinRelation(1, Cor_SchemaSymbol,
                Cor_SchemaSolver);
    temp = new Cor_BuiltinRelation(1, Cor_DelSchemaSymbol,
                Cor_DelSchemaSolver);
                                                                             
    temp = new Cor_BuiltinRelation(1, Cor_BulkLoadSymbol,
                Cor_BulkLoadSolver);

    temp = new Cor_BuiltinRelation(0, Cor_BeginTransSymbol,
                Cor_TransBeginSolver);
    temp = new Cor_BuiltinRelation(0, Cor_CommitTransSymbol,
                Cor_TransCommitSolver);
    temp = new Cor_BuiltinRelation(0, Cor_AbortTransSymbol,
                Cor_TransAbortSolver);

        // These are outdated
    temp = new Cor_BuiltinRelation(1, Cor_StartBulkLoadSymbol,
                Cor_StartBulkLoadSolver);
    temp = new Cor_BuiltinRelation(1, Cor_EndBulkLoadSymbol,
                Cor_EndBulkLoadSolver);
    temp = new Cor_BuiltinSetRelation(-1, Cor_ExtJoinSymbol, Cor_ExtJoinSolver);
}

// *************************************************************************
