// *************************************************************************
//    Defunct.C
// *************************************************************************

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "Defunct.h"

#include "TermList.h"
#include "Print.h"
#include "FunctArgs.h"

#include "Globals.h"
#include "Rule.h"
#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "CArg.h"
#include "Collection.h"

// *************************************************************************

extern CoralStream *outf;
extern int orig_rule_count;

// *************************************************************************

static int var_count = 0;
static int duplicate_var_count = 0;

#define TRACE(x) if( 0 ) { *cerr_strm << x << flush; } else {}

// *************************************************************************

static int add_variable(Cor_Clause *rule, Cor_CArg& name);
static void add_equ_literal(Cor_Clause *rule, const Cor_CArg&, const Cor_CArg&);
static void replace_equ_literal(Cor_Clause *rule, Cor_Literal *lit);
static Cor_CArg *make_new_var(Cor_Clause *rule, const Cor_CArg&);
static void make_distinct_vars(Cor_Clause *rule, Cor_Literal *lit);
static void make_distinct_vars(Cor_Clause *rule);
static void collect_terms(Cor_Clause *rule , TermList *terms);
static void explode(TermList *terms);
static void associate_vars(Cor_Clause *rule, TermList *terms);

#ifdef    NOT_USED
static void freeup(TermList *terms);
#endif

static void reverse(TermList *terms);
static void reverse_list(TermList *terms);
static Cor_CArg *in_termlist(TermList *terms, Cor_CArg *a,
        Cor_Literal * &pred);
static Cor_CArg *in_termlist(TermList **terms, Cor_CArg *a,
        Cor_Literal * &pred);
static void remove_special_pred(Cor_Clause *rule, Cor_Literal *pred);
static void restore_arglist(Cor_Clause *rule, TermList **terms,
        Cor_ArgList &alist, Cor_Literal *p);

// *************************************************************************
void defunctify(Cor_Collection *rules, TermList **terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *r;
    int i;
    int nrules;

    nrules = rules->count();

    // Process rules
    for (i=0; i < nrules; i++) {
        r = rules->get_ith(i);
        // This check is just to ensure that the TermList array
        // has one slot for every rule.
        ASSERT(i < orig_rule_count);

        /*
            Ensure that all non-functor arguments are variables
            by introducing additional equ_(X,Y) literals on the rhs.
            NOTE:: this should actually be handled along with
            functors in an identical fashion. This current approach is
            only a kludge --- PRAVEEN
        */
        make_distinct_vars(r);

        // collect set of all terms with function symbols
        terms[i] = new TermList;
        collect_terms(r, terms[i]);

        // Generate subterms for the set of subterms and associate
        // using forward pointers.
        explode(terms[i]);

        // Reverse the term list, so that nested funcargs are expanded
        // in the right order.
        reverse_list(terms[i]);

        // Do a scan from smallest subterm, and associate with each term
        // a unique variable symbol.
        associate_vars(r, terms[i]);
    }
}

// *************************************************************************
static int add_variable(Cor_Clause *rule, Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg(name);

    if (! rule->var_names)
        rule->var_names = new Cor_CArgStack();

    rule->var_names->push(carg);
    return( rule->var_names->count() - 1 );
}

// *************************************************************************
/*
    Add a literal of the form equ_(old_arg, new_vararg).
*/

static void add_equ_literal(Cor_Clause *rule, const Cor_CArg& arg0,
        const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *tmp2 = new Cor_Literal(Cor_parserStruct, *Cor_EquOpSymbol, 2);
                         
    tmp2->adorn.clear();
    tmp2->args[0] = arg0;
    tmp2->args[1] = arg1;
    tmp2->base_pred = 1;
    rule->preds = (Cor_Literal **) realloc(rule->preds,
        ++(rule->num_preds) * sizeof(Cor_Literal *));
    rule->preds[rule->num_preds - 1] = tmp2;
}

// *************************************************************************
static void replace_equ_literal(Cor_Clause *rule, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    ASSERT(lit != NULL);
    if (! (lit->name().equals(Cor_EquOpSymbol))) {
        // This should not happen
        return;
    }

    for (j = 0; j < rule->head->args.arity(); j++) {
        if (rule->head->args[j] == lit->args[1]) {
            rule->head->args[j] = lit->args[0];
            return;
        }
    }
    
    for (int i = 0; i < rule->num_preds; i++) {
        if (rule->preds[i] == lit) break;
        for (j = 0; j < rule->preds[i]->args.arity(); j++)
            if (rule->preds[i]->args[j] == lit->args[1]) {
                rule->preds[i]->args[j] = lit->args[0];
                return;
            }
    }
}

// *************************************************************************
static Cor_CArg *make_new_var(Cor_Clause *rule, const Cor_CArg& c)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char buf[100];
    sprintf(buf, "V%s__%d", c.symbol(), var_count++);
    Cor_CArg *ret = new Cor_CArg(buf);
    ret->set_var(add_variable(rule,*ret));
    return(ret);
}

// *************************************************************************
static void make_distinct_vars(Cor_Clause *rule, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT( lit != NULL );

    int j = 0;
    Cor_CArg a;
    for (j = 0; j < (lit->args).arity(); j++) {
        a = (lit->args)[j];

        switch (((lit->args)[j]).Kind()) {

            case INT_ARG:
            case DOUBLE_ARG:
            case SYMBOL_ARG:
            case REL_ARG:
            case SUM_GRPING_ARG:
            case PROD_GRPING_ARG:
            case CNT_GRPING_ARG:
            case AVG_GRPING_ARG:
            case SUM_DGRPING_ARG:
            case PROD_DGRPING_ARG:
            case CNT_DGRPING_ARG:
            case AVG_DGRPING_ARG:
            case MIN_GRPING_ARG:
            case MAX_GRPING_ARG:
            case SET_GRPING_ARG:
            case ANY_GRPING_ARG:
            case NO_AGG_GRPING_ARG:
                a = (lit->args)[j];
                (lit->args)[j] = *(make_new_var(rule, Cor_CArg(lit->name())));
                duplicate_var_count++;
                add_equ_literal(rule, a, (lit->args)[j]);
                break;
            case VAR_ARG:
            case DONT_CARE_ARG: {
                int i;
                for (i = j+1; i < (lit->args).arity(); i++) {
                    if ((lit->args)[i].nm_equals((lit->args)[j])) {
                        a = lit->args[i];
                        lit->args[i] = *(make_new_var(rule, lit->name()));
                        duplicate_var_count++;
                        add_equ_literal(rule, a, (lit->args)[i]);
                    }
                    else 
                        break;
                }
            }
            default:
                break;
        }
    }
}

// *************************************************************************
static void make_distinct_vars(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    duplicate_var_count = 0;
    make_distinct_vars(rule, rule->head);
    for (i = 0; i < rule->num_preds; i++) {
        if (! (rule->preds[i]->name().equals(Cor_EquOpSymbol)))
            make_distinct_vars(rule, rule->preds[i]);
    }

    // Ghastly kludge to store the number of extra predicates created in
    // this phase.  old_num_var_names is a field that is not being used,
    // so i'm using it for this purpose -- PRAVEEN
    rule->old_num_var_names = duplicate_var_count;
}

// *************************************************************************
/*
This only collects top level terms that are arguments of predicates in a rule 
*/

static void collect_terms(Cor_Clause *rule , TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    get_terms(rule->head, terms );
    TRACE(("Got terms for head\n"));
    for(i = 0; i < rule->num_preds; i++) {
        get_terms( rule->preds[i], terms );
        TRACE("Got new set of terms for body pred\n" );
    }
    TRACE("finished collecting terms\n");
}

// *************************************************************************
/*
    When exploding a term, we look at each sub term.  If the subterm is 
    already present in the final list, we add a forward pointer from
    the term to the unique copy in the final list.  If it is not present,
    we copy the subterm and introduce it at the end of final list.  We also 
    introduce a forward pointer from the subterm to the copy in the
    final list.  We iteratively end up exploding all subterms since they
    are put at the end of final list and are eventually processed by this
    procedure.
*/

static void explode(TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    TermList *tmp_terms = terms->next;

    for (; tmp_terms; tmp_terms = tmp_terms->next) {
        ASSERT(tmp_terms->func_arg->is_functor());
        int j = 0;
        FOR_EACH_ARG(a, tmp_terms->func_arg->args()) {
            if (a->is_functor()) {
                TermList *tmp = find_arg(terms, a);
                add_arg(terms,a, &(tmp_terms->func_arg->args()), j, tmp);
            }
            j++;
        } END_EACH_ARG
    }
    TRACE("Exploded term list\n");
}

// *************************************************************************
/*
    Associates a new variable with each occurrence of a term in the 
    exploded term list associated with the rule.  
*/

static void associate_vars(Cor_Clause * rule, TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char buf[100];

    for (terms = terms->next; terms; terms = terms->next) {
        if (terms->similar) {
            TermList *tmp = terms->similar;
            while (tmp->similar) tmp = tmp->similar;
            (*((Cor_ArgList *)(terms->arg_list)))[terms->arg_pos]
                    = *(tmp->var_arg);
            terms->var_arg = tmp->var_arg;
            continue;
        }            

        sprintf(buf, "%s__%d", terms->func_arg->functor(), var_count);

        Cor_CArg *var_arg = new Cor_CArg(buf);
        var_arg->set_var(add_variable(rule, *var_arg));
        terms->var_arg = var_arg;
        (*((Cor_ArgList *)(terms->arg_list)))[terms->arg_pos]
                    = *(terms->var_arg);

        // Now add an extra predicate to the rule
        // COR_NOTE :: Forced to use realloc() here instead of new() because
        // alloc_preds() initially used malloc !!! --- PRAVEEN

        sprintf(buf, "func__%d", var_count++); 
        Cor_CArg tmp_arg = Cor_CArg(buf);
        Cor_Literal *tmp2 = new Cor_Literal(Cor_parserStruct,
                tmp_arg, terms->func_arg->arity()+1);
        tmp2->adorn.clear();
        int k;
        for (k=0; k< terms->func_arg->arity();k++)
            tmp2->args[k] = terms->func_arg->args()[k];
        tmp2->args[k] = *(terms->var_arg);
        tmp2->base_pred = 1;
        rule->preds = (Cor_Literal **)realloc(rule->preds,
            ++(rule->num_preds) * sizeof(Cor_Literal *));
        rule->preds[rule->num_preds - 1] = tmp2;
        terms->dummy_pred = tmp2;
    }

    TRACE("Finished associating vars\n");
}

#ifdef    NOT_USED
// *************************************************************************
static void freeup(TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!terms) return;
    while (terms) {
        TermList *tmp = terms->next;
        delete terms;
        terms = tmp;
    }
}

#endif
// *************************************************************************
static void reverse(TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (terms->next == NULL) return;
    reverse(terms->next);
    terms->next->next = terms;
    terms->next = NULL;
}

// *************************************************************************
static void reverse_list(TermList *terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    TermList *tmp = terms->next;
    if (!tmp) return;
    while (tmp->next) tmp = tmp->next;
    reverse(terms->next);
    terms->next = tmp;
}

// *************************************************************************
static Cor_CArg *in_termlist(TermList *terms, Cor_CArg *a,
        Cor_Literal * &pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!terms)
        return NULL;

    for (terms = terms->next; terms; terms= terms->next) {
        if (a->nm_equals(*(terms->var_arg))) {
            pred = terms->dummy_pred;
            return terms->func_arg;
        }
    }
    return NULL;
}

// *************************************************************************
static Cor_CArg *in_termlist(TermList **terms, Cor_CArg *a,
        Cor_Literal * &pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int j=0; j < orig_rule_count; j++) {
         Cor_CArg *tmp_funcarg = in_termlist(terms[j], a, pred);
         if (tmp_funcarg)
             return tmp_funcarg;
    }
    return NULL;
}

// *************************************************************************
static void remove_special_pred(Cor_Clause *rule, Cor_Literal *pred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    for (i = 0; i< rule->num_preds; i++) {
    if ((pred->name() == rule->preds[i]->name()) &&
        pred->arity() == rule->preds[i]->arity()) break;
    }

    // If not found, return ---- this should not happen !!
    if (i == rule->num_preds) return;

    // Copy the adornment in the literal 
    pred->adorn = rule->preds[i]->adorn;

    // Remove the literal from the rule
    for (; i< rule->num_preds; i++) rule->preds[i] = rule->preds[i++];
    (rule->num_preds)--;
}

// *************************************************************************
//    Replace dummy variables by original func args.

static void restore_arglist(Cor_Clause *rule, TermList **terms,
        Cor_ArgList &alist, Cor_Literal *p)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int k = 0;
    Cor_CArg *f_arg;
    FOR_EACH_ARG(a, alist) {
        if (p && p->adorn[k])
            alist[k] = *Cor_TheDontCareArg;

        if (! a->is_var())
            continue;

        // At this point, all args must be of type VAR_ARG

        Cor_Literal *tmp_pred = NULL;
        if ((f_arg = in_termlist(terms, a, tmp_pred))) {
            remove_special_pred(rule, tmp_pred);
            if (!alist[k].equals(*Cor_TheDontCareArg)) {
                alist[k] = *f_arg;
                restore_arglist(rule, terms, f_arg->args(), tmp_pred);
            }
        }
        k++;
    } END_EACH_ARG
}

// *************************************************************************
void functify(Cor_Collection *rules, TermList **terms)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i, j, nrules;
    Cor_Clause *r;

    nrules = rules->count();

      // Process rules 

    for (i=0; i < nrules; i++) {
        r = rules->get_ith(i);

        // Process predicates
        restore_arglist(r, terms, r->head->args, NULL);

        for (j=0; j< r->num_preds; j++)
            restore_arglist(r, terms, r->preds[j]->args, NULL);

        for (j=0; j< r->old_num_var_names; j++)
            replace_equ_literal(r, r->preds[r->num_preds - j -1]);

        r->num_preds -= r->old_num_var_names;
    }
}

// *************************************************************************
